// SpawnParticles(x, y, direction, color)
PORTAL.SpawnParticles = '
    if(global.particles == PARTICLES_OFF) {
        exit;
    }

    var blood;
    repeat(4) {
        blood = instance_create(argument0, argument1, BloodDrop);
        blood.speed = random(6);
        blood.direction = argument2 + random(43) - 22;
        blood.lifetime = 60 + random(51) - 25;

        // Set a white sprite so we can color it properly.
        blood.sprite_index = BloodDropSP;
        blood.image_index = 6;
        blood.color = argument3;
    }
';

// TeleportProjectile(projectile, portal, [don't rotate])
PORTAL.TeleportProjectile = '
    if (argument1.pair == noone) {
        exit;
    }

    if (variable_local_exists("lastPortal") == true) {
        if (lastPortal == argument1.id) {
            exit;
        }
    }
    lastPortal = argument1.pair;

    // Shoot out of the other portal with your current velocity.
    var tmpSpeed, tmpDir;
    tmpSpeed = speed;
    tmpDir = 180 + argument1.image_angle - direction + argument1.pair.image_angle;
    hspeed = lengthdir_x(tmpSpeed, tmpDir);
    vspeed = lengthdir_y(tmpSpeed, tmpDir);

    if (argument2 != true) {
        image_angle = direction;
    }

    x = argument1.pair.x + lengthdir_x(10, tmpDir);
    y = argument1.pair.y + lengthdir_y(10, tmpDir);

    // We are most likely in the wall now, move outside the wall in the direction that the portal is facing,
    // then move contact in the other direction.
    charSetAllSolids();
    with(Obstacle) {
        solid = true;
    }
    with(DropdownPlatform) {
        solid = true;
    }
    move_outside_solid(argument1.pair.image_angle, 50);
    move_contact_solid(argument1.pair.image_angle + 180, 50);
    charUnsetSolids();
    with(DropdownPlatform) {
        solid = false;
    }
';

// distance Raycast(colliders, maxLen, x1, y1, x2, y2) (out hitInstance)
PORTAL.Raycast = '
    var hit, i, len, x1, y1, x2, y2, xm, ym;
    len = argument1;
    x1 = argument2;
    y1 = argument3;
    x2 = argument4;
    y2 = argument5;
    hitInstance = noone;

    while(len > 1) {
        xm=(x1+x2)/2;
        ym=(y1+y2)/2;

        hit = noone;
        for(i = 0; (i < ds_list_size(argument0)) and !hit; i += 1) {
            hit = collision_line(x1, y1, xm, ym, ds_list_find_value(argument0, i), true, true);
        }

        if (hit) {
            x2=xm;
            y2=ym;
            hitInstance = hit;
        } else {
            x1=xm;
            y1=ym;
        }
        len/=2;
    }

    return point_distance(argument2, argument3, xm, ym);
';

// SendPlacePortal(x, y, dir, color)
PORTAL.SendPlacePortal = '
    var portal, _x, _y, _dir;
    _x = round(argument0 * 5);
    _y = round(argument1 * 5);
    _dir = (argument2 + 360) mod 360;

    write_ubyte(PORTAL.sendBuffer, PORTAL.PLACE_PORTAL);

    write_ushort(PORTAL.sendBuffer, _x);
    write_ushort(PORTAL.sendBuffer, _y);
    write_ubyte(PORTAL.sendBuffer, _dir / 2);
    write_ubyte(PORTAL.sendBuffer, color_get_red(argument3));
    write_ubyte(PORTAL.sendBuffer, color_get_green(argument3));
    write_ubyte(PORTAL.sendBuffer, color_get_blue(argument3));

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);

    portal = execute_string(PORTAL.PlacePortal, _x / 5, _y / 5, _dir, argument3);
    return portal;
';

// instance PlacePortal(x, y, dir, color)
PORTAL.PlacePortal = '
    var portal;
    portal = instance_create(argument0, argument1, PORTAL.Portal);
    portal.image_angle = argument2;
    portal.color = argument3;
    portal.alarm[0] = portal.animationTime;

    // Link portals when they are created.
    var complementaryColor;
    complementaryColor = make_color_rgb(255 - color_get_red(portal.color), 255 - color_get_green(portal.color), 255 - color_get_blue(portal.color));
    with(PORTAL.Portal) {
        if (color == complementaryColor) {
            if (instance_exists(pair) == true) {
                pair.image_index = 0;
            }
            image_index = 1;
            portal.image_index = 1;

            pair = portal;
            portal.pair = id;
            break;
        }
    }

    // Link these portals so they are marked on the cursor.
    if (global.myself.object != -1) {
        if (global.myself.object.currentWeapon.object_index == PORTAL.Portalgun) {
            var col;
            col = make_color_rgb(global.myself.object.currentWeapon.colorR, global.myself.object.currentWeapon.colorG, global.myself.object.currentWeapon.colorB);
            if (col == argument3) {
                global.myself.object.currentWeapon.primary = portal;
            } else {
                col = make_color_rgb(255 - global.myself.object.currentWeapon.colorR, 255 - global.myself.object.currentWeapon.colorG, 255 - global.myself.object.currentWeapon.colorB);
                if (col == argument3) {
                    global.myself.object.currentWeapon.secondary = portal;
                }
            }
        }
    }

    return portal;
';

// SendDestroyPortal(portal)
PORTAL.SendDestroyPortal = '
    write_ubyte(PORTAL.sendBuffer, PORTAL.DESTROY_PORTAL);

    write_ubyte(PORTAL.sendBuffer, ds_list_find_index(PORTAL.portals, argument0));

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);

    execute_string(PORTAL.DestroyPortal, ds_list_find_index(PORTAL.portals, argument0));
';

// DestroyPortal(portalId)
PORTAL.DestroyPortal = '
    var portal;
    portal = ds_list_find_value(PORTAL.portals, argument0);
    portal.alarm[1] = portal.animationTime;

    portal.image_index = 0;
    if (instance_exists(portal.pair) == true) {
        if (portal.pair.pair == portal) {
            portal.pair.image_index = 0;
        }
    }
';

// SendPlacePedestal(x, y, color, respawnTime)
PORTAL.SendPlacePedestal = '
    var pedestal;
    write_ubyte(PORTAL.sendBuffer, PORTAL.PLACE_PEDESTAL);

    write_ushort(PORTAL.sendBuffer, argument0 * 5);
    write_ushort(PORTAL.sendBuffer, argument1 * 5);
    write_ubyte(PORTAL.sendBuffer, color_get_red(argument2));
    write_ubyte(PORTAL.sendBuffer, color_get_green(argument2));
    write_ubyte(PORTAL.sendBuffer, color_get_blue(argument2));
    write_ushort(PORTAL.sendBuffer, argument3);

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);

    pedestal = execute_string(PORTAL.PlacePedestal, argument0, argument1, argument2, argument3);
    return pedestal;
';

// instance PlacePedestal(x, y, color, respawnTime)
PORTAL.PlacePedestal = '
    var pedestal;
    pedestal = instance_create(argument0, argument1, PORTAL.Pedestal);
    pedestal.defaultColor = argument2;
    pedestal.color = argument2;
    pedestal.respawnTime = argument3;
    pedestal.ownedByMap = false;

    return pedestal;
';

// SendEquip(player, pedestal, color)
PORTAL.SendEquip = '
    var playerId, pedestalId;
    playerId =  ds_list_find_index(global.players, argument0);
    if (argument1 > 0) {
        pedestalId =  ds_list_find_index(PORTAL.pedestals, argument1);
    } else {
        pedestalId = 255;
    }

    write_ubyte(PORTAL.sendBuffer, PORTAL.EQUIP);

    write_ubyte(PORTAL.sendBuffer, playerId);
    write_ubyte(PORTAL.sendBuffer, pedestalId);
    write_ubyte(PORTAL.sendBuffer, color_get_red(argument2));
    write_ubyte(PORTAL.sendBuffer, color_get_green(argument2));
    write_ubyte(PORTAL.sendBuffer, color_get_blue(argument2));

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);

    execute_string(PORTAL.Equip, playerId, pedestalId, argument2);
';

// Equip(playerId, pedestalId, color)
PORTAL.Equip = '
    var player, pedestal;
    player = ds_list_find_value(global.players, argument0);

    if (player.object == -1) {
        show_error("Giving a portalgun to a dead player GG.", false);
        exit;
    }

    if (argument1 != 255) {
        pedestal = ds_list_find_value(PORTAL.pedestals, argument1);

        if (player.object.currentWeapon.object_index == PORTAL.Portalgun) {
            pedestal.color = make_color_rgb(player.object.currentWeapon.colorR, player.object.currentWeapon.colorG, player.object.currentWeapon.colorB);
        } else {
            pedestal.hasPortalgun = false;
            player.object.sourcePedestal = pedestal;
            if (pedestal.respawnTime > 0) {
                with(pedestal) {
                    alarm[0] = respawnTime / global.delta_factor;
                }
            }
        }
    }

    with(player.object.currentWeapon) {
        instance_destroy();
    }

    global.paramOwner = player.object;
    player.object.currentWeapon = instance_create(x, y, PORTAL.Portalgun);
    global.paramOwner = noone;

    player.object.currentWeapon.colorR = color_get_red(argument2);
    player.object.currentWeapon.colorG = color_get_green(argument2);
    player.object.currentWeapon.colorB = color_get_blue(argument2);
';

// SendUnequip(player, pedestal)
PORTAL.SendUnequip = '
    var playerId, pedestalId;
    playerId = ds_list_find_index(global.players, argument0);
    if (argument1 > 0) {
        pedestalId =  ds_list_find_index(PORTAL.pedestals, argument1);
    } else {
        pedestalId = 255;
    }

    write_ubyte(PORTAL.sendBuffer, PORTAL.UNEQUIP);

    write_ubyte(PORTAL.sendBuffer, playerId);
    write_ubyte(PORTAL.sendBuffer, pedestalId);

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);

    execute_string(PORTAL.Unequip, playerId, pedestalId);
';

// Unequip(playerId, pedestalId)
PORTAL.Unequip = '
    var player, pedestal;
    player = ds_list_find_value(global.players, argument0);

    if (player.object == -1) {
        show_error("Taking a portalgun from a dead player GG.", false);
        exit;
    }

    if (argument1 != 255) {
        pedestal = ds_list_find_value(PORTAL.pedestals, argument1);

        if (player.object.currentWeapon.object_index == PORTAL.Portalgun) {
            player.object.sourcePedestal = noone;
            with(pedestal) {
                color = make_color_rgb(player.object.currentWeapon.colorR, player.object.currentWeapon.colorG, player.object.currentWeapon.colorB);
                alarm[0] = -1;
                hasPortalgun = true;
            }
        } else {
            show_error("Unequipping non-existing portalguns...", false);
            exit;
        }
    }

    with(player.object.currentWeapon) {
        instance_destroy();
    }

    global.paramOwner = player.object;
    player.object.currentWeapon = instance_create(x, y, player.object.weapons[0]);
    global.paramOwner = noone;
';

// SendFullState()
PORTAL.SendFullState = '
    var i, pedestal, portal, cube;

    write_ubyte(PORTAL.sendBuffer, PORTAL.FULL_STATE);
    write_ubyte(PORTAL.sendBuffer, ds_list_size(PORTAL.pedestals));
    for (i = 0; i < ds_list_size(PORTAL.pedestals); i+= 1) {
        pedestal = ds_list_find_value(PORTAL.pedestals, i);
        if (pedestal.ownedByMap == false) {
            write_ushort(PORTAL.sendBuffer, pedestal.x * 5);
            write_ushort(PORTAL.sendBuffer, pedestal.y * 5);
        }

        write_ubyte(PORTAL.sendBuffer, pedestal.hasPortalgun);
        write_ushort(PORTAL.sendBuffer, pedestal.respawnTime);
        write_ushort(PORTAL.sendBuffer, pedestal.alarm[0] * global.delta_factor);
    }

    write_ubyte(PORTAL.sendBuffer, ds_list_size(PORTAL.portals));
    for (i = 0; i < ds_list_size(PORTAL.portals); i+= 1) {
        portal = ds_list_find_value(PORTAL.portals, i);
        write_ushort(PORTAL.sendBuffer, portal.x * 5);
        write_ushort(PORTAL.sendBuffer, portal.y * 5);
        write_ubyte(PORTAL.sendBuffer, portal.image_angle / 2);
        write_ubyte(PORTAL.sendBuffer, color_get_red(portal.color));
        write_ubyte(PORTAL.sendBuffer, color_get_green(portal.color));
        write_ubyte(PORTAL.sendBuffer, color_get_blue(portal.color));
    }

    write_ubyte(PORTAL.sendBuffer, ds_list_size(PORTAL.cubes));
    for (i = 0; i < ds_list_size(PORTAL.cubes); i+= 1) {
        cube = ds_list_find_value(PORTAL.cubes, i);
        write_ubyte(PORTAL.sendBuffer, cube.image_alpha * 255);
        if (cube.image_alpha != 0) {
            write_ushort(PORTAL.sendBuffer, cube.x * 5);
            write_ushort(PORTAL.sendBuffer, cube.y * 5);
        }
    }

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);
';

// ReadFullStata(buffer)
PORTAL.ReadFullState = '
    var i, count, pedestal, portal, cube;

    count = read_ubyte(argument0);
    for (i = 0; i < count; i+= 1) {
        if (ds_list_size(PORTAL.pedestals) > i) {
            pedestal = ds_list_find_value(PORTAL.pedestals, i);
        } else {
            pedestal = instance_create(read_ushort(argument0) / 5, read_ushort(argument0) / 5, PORTAL.Pedestal);
        }

        pedestal.hasPortalgun = read_ubyte(argument0);
        pedestal.respawnTime = read_ushort(argument0);
        pedestal.alarm[0] = read_ushort(argument0) / global.delta_factor;
    }

    count = read_ubyte(argument0);
    for (i = 0; i < count; i+= 1) {
        portal = instance_create(read_ushort(argument0) / 5, read_ushort(argument0) / 5, PORTAL.Portal);
        portal.image_angle = read_ubyte(argument0) * 2;
        portal.color = make_color_rgb(read_ubyte(argument0), read_ubyte(argument0), read_ubyte(argument0));
    }

    var complementaryColor;
    with(PORTAL.Portal) {
        if (pair != noone) {
            continue;
        }
        portal = id;
        complementaryColor = make_color_rgb(255 - color_get_red(portal.color), 255 - color_get_green(portal.color), 255 - color_get_blue(portal.color));
        with(PORTAL.Portal) {
            if (pair != noone || id == portal) {
                continue;
            }

            if (color == complementaryColor) {
                image_index = 1;
                portal.image_index = 1;

                pair = portal;
                portal.pair = id;
                break;
            }
        }
    }

    count = read_ubyte(argument0);
    for (i = 0; i < count; i+= 1) {
        if (ds_list_size(PORTAL.cubes) > i) {
            cube = ds_list_find_value(PORTAL.cubes, i);
        } else {
            cube = instance_create(-999, -999, PORTAL.Cube);
        }

        cube.image_alpha = read_ubyte(argument0) / 255;
        if (cube.image_alpha != 0) {
            cube.x = read_ushort(argument0) / 5;
            cube.y = read_ushort(argument0) / 5;
        } else {
            cube.x = -999;
            cube.y = -999;
        }
    }
';

// SendPickup(player, cube)
PORTAL.SendPickup = '
    var playerId, cubeId;
    playerId = ds_list_find_index(global.players, argument0);
    cubeId =  ds_list_find_index(PORTAL.cubes, argument1);

    write_ubyte(PORTAL.sendBuffer, PORTAL.PICK_UP);

    write_ubyte(PORTAL.sendBuffer, playerId);
    write_ubyte(PORTAL.sendBuffer, cubeId);

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);

    execute_string(PORTAL.Pickup, playerId, cubeId);
';

// Pickup(playerId, cubeId)
PORTAL.Pickup = '
    player = ds_list_find_value(global.players, argument0);
    cube = ds_list_find_value(PORTAL.cubes, argument1);

    if (player.object != -1) {
        if (player.object.currentWeapon.object_index == PORTAL.Portalgun) {
            playsound(player.object.x, player.object.y, PORTAL.ShotSnd);

            if (cube.target != noone) {
                cube.target.pressed = false;
                cube.target = noone;
            }

            player.object.currentWeapon.holding = cube;
            cube.held = true;
            cube.alarm[10] = 2;

            // We move cubes away so we dont mess up collisions.
            if (cube.object_index == PORTAL.Cube) {
                cube.x = -999;
                cube.y = -999;
                cube.image_alpha = 0;
            }
        }
    }
';

// SendDrop(player, cube)
PORTAL.SendDrop = '
    var playerId, cubeId;
    playerId = ds_list_find_index(global.players, argument0);
    cubeId =  ds_list_find_index(PORTAL.cubes, argument1);

    write_ubyte(PORTAL.sendBuffer, PORTAL.DROP);

    write_ubyte(PORTAL.sendBuffer, playerId);
    write_ubyte(PORTAL.sendBuffer, cubeId);

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);

    execute_string(PORTAL.Drop, playerId, cubeId);
';

// Drop(playerId, cubeId)
PORTAL.Drop = '
    player = ds_list_find_value(global.players, argument0);
    cube = ds_list_find_value(PORTAL.cubes, argument1);

    if (player.object != -1) {
        if (player.object.currentWeapon.object_index == PORTAL.Portalgun) {
            playsound(player.object.x, player.object.y, PORTAL.ShotSnd);
            player.object.currentWeapon.holding = noone;
            cube.x = player.object.currentWeapon.x + player.object.currentWeapon.xoffset * player.object.currentWeapon.image_xscale + lengthdir_x(45, player.object.aimDirection);
            cube.y = player.object.currentWeapon.y + player.object.currentWeapon.yoffset + lengthdir_y(45, player.object.aimDirection) + player.object.equipmentOffset;
            cube.vspeed = player.object.vspeed;
            cube.hspeed = player.object.hspeed;
            cube.image_alpha = 1;
            cube.held = false;
        }
    }
';

// cube SendSyncCube(cube)
PORTAL.SendSyncCube = '
    var cubeId, _x, _y;
    cubeId = ds_list_find_index(PORTAL.cubes, argument0);
    _x = round(argument0.x * 5);
    _y = round(argument0.y * 5);

    switch(argument0.object_index) {
        case PORTAL.Cube:
            write_ubyte(PORTAL.sendBuffer, PORTAL.SYNC_CUBE);
            break;
        case PORTAL.Turret:
            write_ubyte(PORTAL.sendBuffer, PORTAL.SYNC_TURRET);
            break;
        default:
            show_message("Error syncing " + argument0.object_index);
    }

    write_ubyte(PORTAL.sendBuffer, cubeId);
    if (argument0.image_alpha == 0) {
        write_ubyte(PORTAL.sendBuffer, 0);
        _x = -999 * 5;
        _y = -999 * 5;
    } else {
        write_ubyte(PORTAL.sendBuffer, argument0.image_alpha * 255);
        write_ushort(PORTAL.sendBuffer, _x);
        write_ushort(PORTAL.sendBuffer, _y);
    }

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);

    execute_string(PORTAL.SyncCube, cubeId, argument0.image_alpha, _x / 5, _y / 5, argument0.object_index);
';

// cube SyncCube(cubeId, image_alpha, x, y, object)
PORTAL.SyncCube = '
    var cube, _x, _y;
    cube = ds_list_find_value(PORTAL.cubes, argument0);

    // Lets abuse this funcion to also spawn new cubes.
    if (cube == 0) {
        cube = instance_create(argument2, argument3, argument4);
        cube.image_alpha = argument1;
        _x = argument2;
        _y = argument3;
    } else {
        _x = cube.x;
        _y = cube.y;
        cube.x = argument2;
        cube.y = argument3;
        cube.image_alpha = argument1;
    }

    if (cube.target != noone) {
        cube.target.pressed = false;
    }

    if (argument1 == 0) {
        var fizzle;
        fizzle = instance_create(_x, _y, PORTAL.FizzleCube);
        with(fizzle) {
            sprite_index = cube.sprite_index;
            image_index = cube.image_index;
            image_xscale = cube.image_xscale;
            motion_add(cube.direction, cube.speed);
        }
        cube.held = false;
    } else if (cube.object_index == PORTAL.Cube) {
        cube.target = collision_line(cube.x, cube.y, cube.x, cube.y + 15 , PORTAL.Plate, true, false);
        if (cube.target != noone) {
            if (cube.target.pressed == false) {
                playsound(cube.target.x, cube.target.y, PORTAL.ButtonSnd);
            }
            cube.target.pressed = true;
            cube.target.lastCube = cube;
        }
    }

    cube.vspeed = 0;
    cube.hspeed = 0;
    cube.synced = true;

    return cube;
';

PORTAL.SendButtonPress = '
    var buttonId;
    buttonId = ds_list_find_index(PORTAL.buttons, argument0);

    write_ubyte(PORTAL.sendBuffer, PORTAL.BUTTON_PRESS);

    write_ubyte(PORTAL.sendBuffer, buttonId);

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);

    execute_string(PORTAL.ButtonPress, buttonId);
';

// ButtonPress(buttonId);
PORTAL.ButtonPress = '
    var button;
    button = ds_list_find_value(PORTAL.buttons, argument0);

    with(button) {
        playsound(x, y, PORTAL.ButtonSnd);
        pressed = true;
        alarm[0] = timer / global.delta_factor;
    }
';

// SendRightClick(player)
PORTAL.SendRightClick = '
    var playerId;
    playerId = ds_list_find_index(global.players, argument0);

    write_ubyte(PORTAL.sendBuffer, PORTAL.RIGHT_CLICK);

    write_ubyte(PORTAL.sendBuffer, playerId);

    PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
    buffer_clear(PORTAL.sendBuffer);

    execute_string(PORTAL.RightClick, playerId);
';

// RightClick(playerId)
PORTAL.RightClick = '
    var player;
    player = ds_list_find_value(global.players, argument0);

    if (player.object != -1) {
        if (player.object.currentWeapon.object_index == PORTAL.Portalgun) {
            with(player.object.currentWeapon) {
                event_user(2);
            }
        }
    }
';
